/*
 * agpgart.c: Main file for agpgart driver for Solaris
 * Written by Philip Brown : http://www.bolthole.com/solaris/
 *
 * see LICENSE file
 * @(#) agp_i440bx.c 1.14@(#)
 *
 * The solaris-specific stuff, and 100% generic stuff, is kept mostly in
 * agpgart.c.
 * THIS file targets the AGP hardware on an 
 * intel 440BX motherboard AGP chipset (intel 82443BX PCI device)
 * Specs for the chip can be found at
 *  http://developer.intel.com/design/chipsets/datashts/29063301.pdf
 * Despite the name of the file, it contains the 82443BX details
 *
 * To add support for another AGP chipset, copy this file to
 * agp_yourchipset.c, adjust the i440bx names, and then start
 * tweaking the routines.
 * Basically, you'll want to implement all the
 * ioctl_i440bx_XXXXX()  routines as appropriate for your chipset
 * Specifically,
 *	ioctl_i440bx_acquire
 *	ioctl_i440bx_release
 *	ioctl_i440bx_alloc
 *	ioctl_i440bx_dealloc
 *	ioctl_i440bx_bind
 *	ioctl_i440bx_unbind
 *	ioctl_i440bx_info
 *	ioctl_i440bx_protect
 *	ioctl_i440bx_reserve
 *	ioctl_i440bx_setup
 *
 * Nothing else in this file is referenced from outside it.
 *
 *  NOTE1: Read all the comments for each ioctl handler
 *  NOTE2: Notice how paranoid I am about handling dumb user error.
 *         Do likewise.
 *  NOTE3:   DO NOT MESS WITH MUTEXES!!!
 *         The mainline controls entry to all routines in this file.
 *         It already grabs instance_mutex, so any routine in here
 *         is effectively the single thread running for the
 *         whole driver.
 *  NOTE4: Some day, this file will probably metamorphose into a more
 *         generic "intel agp" file. That time is not now.
 */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/types.h>

/* THESE MUST BE THE LAST of the system includes */
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "agpgart.h"
#include "agpdrv.h"




static ddi_device_acc_attr_t dev_access={
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};


/************************************************************************\
 * Intel-specific defines go in HERE, not the agp_i440bx.h header file. *
 *  Notice that we do not INCLUDE that file !                           *
\************************************************************************/


 
#define GATT_MAGIC 0x17	/* "valid" marker for intel GATT table entry */

/* Defines for register offsets in the PCI config space     */
/*   http://developer.intel.com/design/intarch/techinfo/440BX/HostPCI.htm */

#define CONF_APBASE	0x10	/* aperture base config register.
				 * Wierd stuff. See APSIZE instead, and
				 * maybe consider this read-only.
				 * Dont forget about NBXCFG_AP_ENABLE
				 */

#define CONF_NBXCFG	0x50	/* 4byte*/
#define CONF_CMDERR	0x90	/* 1 byte */
#define CONF_ERRSTATUS	0x91	/* 2 byte */
#define CONF_AGPSTATUS	0xa4
#define CONF_AGPCAPID	0xa0	/* AGP Capability reg */
#define CONF_AGPCMD	0xa8
#define CONF_AGPCTRL	0xb0
#define CONF_APSIZE	0xb4	/* "aperture size" (1 byte, encoded)*/
				/* Defaults to 4mb, but is WRITABLE.
				 * If changed, I should probably
				 * unmap and remap agpgart_regs if
				 * I think I can mmap them as RAM.
				 * (maybe I cant do that anyway!)
				 */
#define CONF_ATTBASE	0xb8	/* "aperture translation table"
				 * 'base address'.
				 * We allocate a GATT table and point
				 * this register to it.
				 * Note that the only valid bits are:
				 * 0xff ff f0 00
				 * Which is another way of saying,
				 * "must start on 4k page boundary".
				 */

/* Defines that are specifically used with CONF_NBXCFG register */
#define NBXCFG_AP_ENABLE		(1<<9)	/* enable aperture */
						/* only use after GATT done */
#define NBXCFG_AP_PCI_DISABLE	(1<<10) /* disable PCI access to aperture*/
#define NBXCFG_AP_AGP_ENABLE	(1<<11)	/* enable AGP->PCI access */

/* Defines that are specifically used with CONF_AGPSTATUS register */
/* Note how some of these mirror bits for CMD_XXXX */
#define STATUS_SBALLOWED	(1<<9)
#define STATUS_SPEEDBITS	0x03	/* note match to CMD_SPEEDMASK */
#define STATUS_RQBITS	(0xff000000) /* do (x>>24) on result */

/* Defines that are specifically used with CONF_AGPCMD register */
#define CMD_SBENABLE		(1<<9)	 /* "Side Band enable" */
#define CMD_AGPENABLE		(1<<8)	 /* Allow AGP operations */
#define CMD_AGPSPEED0		0x0	/* not sure this should ever be used*/
#define CMD_AGPSPEED1		0x1	/* AGPx1 */
#define CMD_AGPSPEED2		0x2	/* AGPx2 */
#define CMD_SPEEDBITS		0x3
#define	CMD_RQBITS		0xff000000	/* Read Queue */
						/* see also STATUS_RQBITS */

/* Defines that are specifically used with CONF_AGPCTRL register */
#define CTRL_AGPSYNC		(1<<13)
#define CTRL_INTELMAGIC		(1<<9)	/* undocumented magic */
#define CTRL_TLBENABLE		(1<<7)	/* Translation Lookaside Buffer*/
 					/* clear bit to flush buffer */

/*************************************************************************/
/* Start of local utility functions, and private hardware-tweaking funcs */
/*************************************************************************/

extern void agpgart_debugprint(const char *,...);


/* Quickie hack to set softstate->aperture_size.
 * This is the only routine that goes out and read the HARDWARE
 * to find out the aperture size.
 *
 * Called by ioctl_i440bx_acquire
 */
static void i440bx_read_apsize(struct agpgart_softstate *softstate){
	uint8_t apsize;

	if(softstate==NULL){
		printf("ERR: read_apsize passed NULL");
		return;
	}
	apsize=pci_config_get8(softstate->pcihandle,CONF_APSIZE);
	agpgart_debugprint("apsize WAS %x: forcing it to be 0x3c\n",
			   apsize);

	apsize=0x3c;

	agpgart_debugprint("DEBUG: ATTBASE=%x\n",
	       pci_config_get32(softstate->pcihandle,CONF_ATTBASE));

	softstate->aperture_size=agp_apsize_to_bytes(apsize);
	agpgart_debugprint("DEBUG: apsize reg=0x%x => memsize=0x%x\n",apsize,
	       softstate->aperture_size);
}


/*
 * Turn on AGP aperture mapping, set AGP speed, etc.
 * YOU MUST HAVE SET UP THE GATT TABLE before calling this routine!!!
 *
 * return 0 on success, or an appropriate errno value otherwise
 */
static int i440bx_enable_agp(struct agpgart_softstate *softstate){
	uint32_t temp32, command;

	/* This comes out to 0x2280.
	 * 0x2280 is magic bits copied  from freebsd agp_intel.c
	 *
	 * dont know what the 0x2000 does, thats the "magic"
	 */
	temp32=CTRL_INTELMAGIC | CTRL_AGPSYNC | CTRL_TLBENABLE;
	pci_config_put32(softstate->pcihandle,CONF_AGPCTRL,temp32);

	temp32=pci_config_get32(softstate->pcihandle,CONF_NBXCFG);
	temp32 |= NBXCFG_AP_ENABLE;
	temp32 &= ~(NBXCFG_AP_PCI_DISABLE);
	pci_config_put32(softstate->pcihandle,CONF_NBXCFG,temp32);



	/* XXX in theory, should limit stuff like sideband enable,
	 * and speed, etc. based on what user selected, and
	 * what the CARD supports.
	 * But we cant see the card from this driver since suns driver
	 * has bound to it. Oh well.
	 * So just enable AGP without sideband support.
	 * On the brighter side, the AGP bridge IS supposed to tell us
	 * the agp speed of the card plugged into it.
	 */

	command=CMD_AGPSPEED1 | CMD_AGPENABLE;
	/* command|=CMD_SBENABLE */
	command |= (temp32 & STATUS_RQBITS);
	command |= (temp32 & STATUS_SPEEDBITS);

	/* NOTE: setting CMD_SPEEDBITS==0x3 is technically illegal,
	 * according to intel spec on the 82443
	 */
	if((command & CMD_SPEEDBITS) == 0x3){
		agpgart_debugprint("WARNING: Resetting speedbits to 0x02\n");
		command &= 0x2;
	}
	pci_config_put32(softstate->pcihandle,CONF_AGPCMD, command);

	return 0;
	
}

/* Adjust the AGP hardware: disable the mapping, and pretty much
 * everything else on the AGP chipset.
 *
 * Dont forget to free the GATT table *AFTER* calling this.
 */
static void i440bx_disable_agp(struct agpgart_softstate *softstate){
	uint32_t temp32;

	pci_config_put32(softstate->pcihandle,CONF_AGPCMD, 0);

	temp32=pci_config_get32(softstate->pcihandle,CONF_NBXCFG);
	temp32 &= ~(NBXCFG_AP_ENABLE);
	pci_config_put32(softstate->pcihandle,CONF_NBXCFG, temp32);

	pci_config_put32(softstate->pcihandle,CONF_ATTBASE, 0);

}

/**********************************************************************/
/**********************************************************************/
/* GATT section                                                       */
/*   (Graphics Aperture Translation Table)                            */
/*   ("Graphics Aperture" is also known as                            */
/*            "AGP DRAM Graphics Aperture"                            */
/* We have to program a chunk of main memory like a scatter/gather    */
/* table, so that the AGP card knows what part of its virtual AGP     */
/*   "aperture" points to what point in real memory                   */
/* We build the table, then point CONF_ATTBASE to the table           */
/**********************************************************************/
/**********************************************************************/


/* utility for i440bx_add_to_gatt(), for error cleanup.
 * Also used by i440bx_alloc_gatt_table().
 * Zero out a range in the gatt table, that covers the agp aperture
 * address range specified.
 */
static void clear_gatt_pages(struct agpgart_softstate *softstate,
			     uint32_t agp_offset, uint32_t length){
	uint32_t *gatt_entry;
	gatt_entry=(uint32_t*)(softstate->GATT+(AGP_BYTESTOPAGES(agp_offset)));

	if(agp_offset+length >softstate->aperture_size){
		agpgart_debugprint("clear_gatt_pages: ERROR: length too big\n");
		return;
	}

	while(length>0){
		ddi_put32(softstate->GATTacc, gatt_entry,0);
		gatt_entry++;
		length-=AGP_PAGE_SIZE;
	}
	ddi_dma_sync(softstate->GATThandle, 0, 0, DDI_DMA_SYNC_FORDEV);

}

/* Utility routine for add_to_gatt.
 * Revalidate dma_handle and cookie for given keyentry.
 * return 0 on okay, 1 on failiure to bind
 */
static int i440_recookie(struct keytable_ent *keyentry)
{
	int bind_val;
	bind_val=ddi_dma_addr_bind_handle(keyentry->dma_handle,
				    NULL,
				    keyentry->kern_addr,
				    keyentry->length,
				    DDI_DMA_RDWR,
				    DDI_DMA_SLEEP,NULL,
				    &keyentry->dma_cookie,
				    &keyentry->num_dma_cookies);
	if(bind_val != DDI_DMA_MAPPED){
		return 1;
	}
	return 0;
}

/*
 * add_to_gatt():
 * Given a physical address, that MUST BE PAGE-ALIGNED,
 * update our GATT table so that the specified physical address
 * gets used as the GATT address, for the given length
 * (which must ALSO be whole pages)
 *
 * We take care of the dma_sync here.
 *
 * Return 0 on okay, EINVAL on "you screwed up", or
 *  return EFAULT on "spot already taken by something"
 *
 * MUTEXES: none. The caller should probably already hold instance_mutex
 *
 * NOTE: This does NOT set the agp_offset field in the keytable.
 *   The caller is expected to do that.
 *
 * Called by AGPIOC_BIND handler.
 */
static int i440bx_add_to_gatt(struct agpgart_softstate *softstate,
		    struct keytable_ent *keyent,uint32_t gatt_offset)
{
	uint32_t *gatt_entry;
	uint32_t tempoffset;
	int num_cookies;

	agpgart_debugprint("i440bx_add_to_gatt called, offset=%d\n",
		gatt_offset);

	if(keyent->num_dma_cookies==0){
		if(keyent->am_bound!=0){
			cmn_err(CE_WARN,
				"i440bx_add_to_gatt: key already bound\n");
			return EINVAL;
			
		}

		/* Otherwise... this key has previously been used.
		 * Trouble is, cookies can only be "eaten" once.
		 * So we have to rebind the handle. 
		 */
		if(i440_recookie(keyent)!=0){
			cmn_err(CE_WARN,
				"i440bx_add_to_gatt: rebind failed\n");
			return EINVAL;
		}
	}
	num_cookies=keyent->num_dma_cookies;



	if(AGP_NONALIGNED(gatt_offset)){
		agpgart_debugprint("add_to_gatt: gatt_offset not aligned\n");
		return EINVAL;
	}
	if(AGP_NONALIGNED(keyent->length)){
		agpgart_debugprint("add_to_gatt: key length not whole pages\n");
		return EINVAL;
	}
	if((gatt_offset + keyent->length) >softstate->aperture_size){
		agpgart_debugprint("add_to_gatt: length overflowed aperture\n");
		agpgart_debugprint("offset=%d, len=%d, aperture=%d\n",
				   gatt_offset, keyent->length,
				   softstate->aperture_size);
		return EINVAL;
	}
	gatt_entry=(uint32_t*)(softstate->GATT+(AGP_BYTESTOPAGES(gatt_offset)));

	/* Quick check to see if there is GATT conflict */

	
	for(tempoffset=gatt_offset; tempoffset < keyent->length;
	    tempoffset+=AGP_PAGE_SIZE)
	{
		uint32_t *entryptr=(uint32_t*)softstate->GATT;
		int pagenum=AGP_BYTESTOPAGES(tempoffset);

		if(ddi_get32(softstate->GATTacc,&entryptr[pagenum]) !=0){
			cmn_err(CE_WARN,
				"i440bx_add_to_gatt: entry at page %x exists\n",
					pagenum	);
			return (EINVAL);
		}
	}

	/* Initial collection of base addresses complete.
	 * Once we are here, either the key stuff is valid, or invalid.
	 * Either we dont want to use it, or we successfully bind it, and
	 * cannot do so again. So clear the cookie count in keyentry
	 */
	keyent->num_dma_cookies=0;

	agpgart_debugprint("  DEBUG: add_to_gatt: Cookie 0 start addr=%x, cookie count=%d\n",
			   keyent->dma_cookie.dmac_address,num_cookies);

	/* Now we have to walk through the dma "cookies", to find out
	 * where the fragments of allocated memory are, and bind each
	 * page into the GATT table.
	 */
	while(num_cookies>0){
		uint32_t ck_startaddr, ck_length, ck_end;
		ck_startaddr=keyent->dma_cookie.dmac_address;
		ck_length=keyent->dma_cookie.dmac_size;

		if(AGP_NONALIGNED(ck_startaddr)){
			agpgart_debugprint("cookie non-pagealigned!\n");
			clear_gatt_pages(softstate,keyent->agp_offset,
					 keyent->length);
			return EIO;
		}
		if(AGP_NONALIGNED(ck_length)){
			agpgart_debugprint("cookie length non-pagealigned!\n");
			clear_gatt_pages(softstate,keyent->agp_offset,
					 keyent->length);
			return EIO;
		}

		ck_end=ck_startaddr+ck_length;
		while(ck_startaddr < ck_end){
			ddi_put32(softstate->GATTacc, gatt_entry,
				  ck_startaddr|GATT_MAGIC);
			ck_startaddr+=AGP_PAGE_SIZE;
			gatt_entry++;
			  
		}

		num_cookies--;
		if(num_cookies>0){
			ddi_dma_nextcookie(keyent->dma_handle,
					   &keyent->dma_cookie);
		}
	}


	ddi_dma_sync(softstate->GATThandle, 0, 0, DDI_DMA_SYNC_FORDEV);
	return 0;	
}

/* Remove a particular chunk of GATT aperture from the mapping.
 * Eg: "invalidate AGP addressing between 0x4000 and 0x8000"
 * gatt_addr must be page-aligned.
 * length must be a size of whole pages.
 *
 * This ONLY adjusts the gatt table. The caller has to
 * deal with the keytable, etc.
 * It is assumed "the caller" is the AGPIOC_UNBIND handler.
 *
 * Return 0 on ok, EINVAL on bad align, or EFAULT on
 * "it was already invalid"
 */
static int i440bx_remove_from_gatt(struct agpgart_softstate *softstate,
			 uint32_t gatt_addr, size_t length)
{
	uint32_t *gatt_entry;
	int pagecount=AGP_BYTESTOPAGES(length);

	if(gatt_addr +length >softstate->aperture_size){
		agpgart_debugprint("remove_from_gatt: val too large\n");
		return EINVAL;
	}

	if(AGP_NONALIGNED(gatt_addr)){
		agpgart_debugprint("remove_from_gatt: gatt_addr not aligned\n");
		return EINVAL;
	}
	if(AGP_NONALIGNED(length)){
		agpgart_debugprint("remove_from_gatt: length not whole pages\n");
		return EINVAL;
	}
	gatt_entry=(uint32_t*)softstate->GATT;
	gatt_entry += (AGP_BYTESTOPAGES(gatt_addr));
	while(pagecount-- >0){
		ddi_put32(softstate->GATTacc, gatt_entry, 0);
		gatt_entry += 1;
	}

	ddi_dma_sync(softstate->GATThandle, 0, 0, DDI_DMA_SYNC_FORDEV);
	return 0;
}


/* Called by AGPIOC_RELEASE handler only.
 * MUTEXES: none. Caller should probably hold instance_mutex
 */
static void i440bx_free_gatt_table(struct agpgart_softstate *softstate){
	if(softstate->GATT ==NULL){
		return;
	}
	if(softstate->GATThandle != NULL){
		if(softstate->GATTphysaddr!=NULL){
			ddi_dma_unbind_handle(softstate->GATThandle);
		}
		ddi_dma_free_handle(&softstate->GATThandle);
	}
	if(softstate->GATTacc!=NULL){
		ddi_dma_mem_free(&softstate->GATTacc);
	}
	softstate->GATT=NULL;
	softstate->GATTphysaddr=NULL;
	softstate->GATThandle=NULL;
	softstate->GATTacc=NULL;

}


/* "GATT table" = Graphics Aperture Translation Table
 * This is called by the AGPIOC_ACQUIRE handler
 *
 * We read softstate->aperture_size to figure out how large a GATT
 * table to allocate.
 *
 * return 0 on okay, 1 on fail
 */
static int i440bx_alloc_gatt_table(struct agpgart_softstate *softstate){

	ddi_dma_cookie_t cookie;
	int num_pages,num_cookies, bind_val;
	size_t tlength;

	agpgart_debugprint("alloc_gatt: aperture_size=%d\n",
			   softstate->aperture_size);

	num_pages=AGP_BYTESTOPAGES(softstate->aperture_size);
	
	tlength=num_pages*(sizeof(uint32_t));

	agpgart_debugprint("alloc_gatt: num_pages=%d, tlength=%d\n",
			   num_pages, tlength);

	/* SHOULD do a check for size_type of "FIXED SIZE"*/
	if(ddi_dma_alloc_handle(softstate->dip,
			     &dev_dma_attr,DDI_DMA_SLEEP,NULL,
			     &softstate->GATThandle)!=DDI_SUCCESS){
		cmn_err(CE_WARN,"ddi_dma_alloc_handle for GATT failed");
		return 1;
	}
	if(ddi_dma_mem_alloc(softstate->GATThandle,
			  tlength, &dev_access, DDI_DMA_CONSISTENT,
			  NULL,NULL,
			  &softstate->GATT,
			  &softstate->GATTlength,
			  &softstate->GATTacc) !=DDI_SUCCESS){

		cmn_err(CE_WARN,"ddi_dma_mem_alloc for GATT failed");
		cmn_err(CE_WARN,"  num_pages=%d,tlength=%d",num_pages,tlength);

		ddi_dma_free_handle(&softstate->GATThandle);
		return 1;
	}
	bind_val=ddi_dma_addr_bind_handle(softstate->GATThandle,
				    NULL,
				    softstate->GATT,softstate->GATTlength,
				    DDI_DMA_CONSISTENT, DDI_DMA_SLEEP, NULL,
				    &cookie, &num_cookies);

	softstate->GATTphysaddr=cookie.dmac_address;

	if(AGP_NONALIGNED(softstate->GATTphysaddr)){
		cmn_err(CE_WARN,"alloc_gatt_table: mem not pagealigned (0x%x)",
			softstate->GATTphysaddr);

		i440bx_free_gatt_table(softstate);
		return 1;
	}

	if((bind_val != DDI_DMA_MAPPED) || (num_cookies != 1)){
		agpgart_debugprint("ERROR: addr_bind_handle failed for gatt: val=%d, cookies=%d\n",
			   bind_val, num_cookies);
		i440bx_free_gatt_table(softstate);
		return 1;
	}

	clear_gatt_pages(softstate,0,softstate->aperture_size);
	agpgart_debugprint("alloc_gatt: complete!\n");
	return 0;
}

/* Once the GATT table is allocated, tell the hardware where to find it.
 * I almost feel like I should have this be combined into the alloc routine,
 * but it gives me warm fuzzies to have separate alloc and install
 * routines.
 * Plus, when and if I get around to supporting multiple intel chipsets,
 * it will probably be prettier this way.
 */
static void i440bx_install_gatt_table(struct agpgart_softstate *softstate){
	pci_config_put32(softstate->pcihandle, CONF_ATTBASE,
			 softstate->GATTphysaddr);
}


/* Extra paranoia: If the hardware somehow accidentally gets left on after
 * we are done, might as well let it know the data isnt valid any more.
 * Call this before calling free_gatt_table.
 */
static void i440bx_clear_gatt_table(struct agpgart_softstate *softstate){
	clear_gatt_pages(softstate,0,softstate->aperture_size);

}


/**********************************************************************/
/*            End of GATT section                                     */
/**********************************************************************/



/**********************************************************************/
/**********************************************************************/
/**********************************************************************/


/* Handles AGPIOC_ACQUIRE
 * This routine is the first ioctl the user must make to even talk
 * to us.
 * It is responsible for allocating the GATT table,
 * reading the current value for softstate->aperture_size,
 * and initializing the AGP hardware.
 *
 * If it returns non-zero, then the _release() routine will not get called.
 *   (because am_enabled will not get set)
 * So it MUST CLEAN UP AFTER ITSELF on fail!!!
 *
 * MUTEXES:
 *    uses instance_mutex
 *
 * Caller may use return value as return for ioctl(9e)
 */
int ioctl_i440bx_acquire(struct agpgart_softstate *softstate, void *arg,
		       int flags)
{
	int val=0;

	
	/* linux version arbitrates between multiple programs and/or
	 * kernel trying to acquire dev.
	 * Forget that, it's done implicitly by the
	 * open call.
	 *
	 * But we DO want to turn on AGP mapping here.
	 * We turn it off in the _release hook
	 */
	if(softstate->GATT!=NULL){
		agpgart_debugprint("agpgart: ioctl_acquire already called\n");
		return EINVAL;

	}
	/* Paranoia check */
	{
		uint32_t temp32, magic=0x1f000200;
		temp32=pci_config_get32(softstate->pcihandle,CONF_AGPSTATUS);

		if((temp32 & magic) != magic){
			agpgart_debugprint("ERR: failed status reg magic test\n");
			return EFAULT;
		} else {
			agpgart_debugprint("ioctl_i440bx_acquire: passed magic test\n");
		}
	}

	if(softstate->aperture_size==0){
		i440bx_read_apsize(softstate);
	}

	val=i440bx_alloc_gatt_table(softstate);

	if(val!=0){
		cmn_err(CE_WARN,"i440bx_alloc_gatt_table failed. ACQUIRE failed.");
		return val;
	}

	i440bx_install_gatt_table(softstate);

	val=i440bx_enable_agp(softstate);
	if(val!=0){
		cmn_err(CE_WARN,"enable_agp returned FAIL. ACQUIRE failed");
		i440bx_free_gatt_table(softstate);
		return val;
	}

	softstate->am_enabled=1;
	return 0;
}

/* convenience routine called by ioctl_i440bx_release() */
static void i440bx_unbind_allkeys(struct agpgart_softstate *softstate){
	int key;
	struct keytable_ent *keyentry;

	agpgart_debugprint("i440bx_unbind_allkeys called");

	for(key=0; key<AGP_MAXKEYS; key++){
		keyentry=agp_get_keyent(softstate,key);
		if(keyentry==NULL){
			continue;
		}
		if(keyentry->am_bound==0){
			continue;
		}
		agpgart_debugprint("unbind_allkeys: unbinding %d\n",key);
		/* Copy what ioctl_i440bx_unbind() does */
		i440bx_remove_from_gatt(softstate,
			       keyentry->agp_offset,
			       keyentry->length);
		keyentry->agp_offset=0;
		keyentry->am_bound=0;
	}
	agpgart_debugprint("i440bx_unbind_allkeys done");
}


/* Handles AGPIOC_RELEASE.
 * 
 * Serves as cleanup for userspace allocations, AND
 * method of deactivating AGP if driver is removed.
 * - Will free up any and all keys.
 * - Will deactivate AGP mappings
 * - Will free up GATT table.
 *
 * Caller may use return value as return for ioctl(9e)
 *
 */
int ioctl_i440bx_release(struct agpgart_softstate *softstate, void *arg,
		       int flags){

	agpgart_debugprint("i440bx_release called\n");

	i440bx_unbind_allkeys(softstate);

	agp_del_allkeys(softstate);

	i440bx_clear_gatt_table(softstate);
	i440bx_free_gatt_table(softstate);
	i440bx_disable_agp(softstate);

	softstate->am_enabled=0;

	return 0;
}

/* AGPIOC_RESERVE */
/* What the heck do I do here??? */
/* Caller may use return value as return for ioctl(9e) */
int ioctl_i440bx_reserve(struct agpgart_softstate *softstate, void *arg,
		       int flags){
	/* linux version arbitrates between multiple programs and/or
	 * kernel trying to acquire dev.
	 * Forget that, it's done implicitly by the
	 * open call
	 */
	return 0;
}
/* AGPIOC_PROTECT */
/* What the heck do I do here??? */
/* Caller may use return value as return for ioctl(9e) */
int ioctl_i440bx_protect(struct agpgart_softstate *softstate, void *arg,
		       int flags){
	return 0;
}

/* handles AGPIOC_INFO */
/* Caller may use return value as return for ioctl(9e) */
int ioctl_i440bx_info(struct agpgart_softstate *softstate, void *arg,
		      int flags){
	agp_info infostruct;

	infostruct.version.major=VERSION_MAJOR;
	infostruct.version.minor=VERSION_MINOR;

	/* They have really stupid order for this, but I will just copy
	 * the stupidity. Vendor should be MSB!
	 */
	infostruct.bridge_id=
		(pci_config_get16(softstate->pcihandle,2) <<16) | /* devID*/
		pci_config_get16(softstate->pcihandle,0) /* vendor */ ;

	infostruct.agp_mode=0; /* doesnt seem to do much? */

	infostruct.aper_base=pci_config_get32(softstate->pcihandle, CONF_APBASE)
			     & 0xfffffff0;
	infostruct.aper_size=(AGP_BYTESTOMEGS(softstate->aperture_size));

	/* Anything that uses the following info is screwed under
	 * solaris anyway. I'm making vague guesses as to appropriate
	 * values.
	 */
	infostruct.pg_total=AGP_BYTESTOPAGES(softstate->aperture_size);
	infostruct.pg_system=infostruct.pg_total;
	infostruct.pg_used=agp_count_keypages(softstate);

	if(ddi_copyout(&infostruct, arg,
		       sizeof(agp_info), flags) !=0){
		return (EFAULT);
	}

	return 0;
}


/* handles AGPIOC_SETUP
 * We ignore mode for now. Otherwise,
 * would set AGP mode here.
 * Also, fill out anything in soft state that is
 * hardware-specific.
 *
 * Caller may use return value as return for ioctl(9e)
 */
int ioctl_i440bx_setup(struct agpgart_softstate *softstate, void *arg,
		       int flags){
	size_t pagecount;
	pagecount=AGP_BYTESTOPAGES(softstate->aperture_size);

	/* This is the big stuff:
	 * - Sets AGP modes in the hardware
	 * - points hardware to gatt table
	 * - sets Read Queue depth to minimum(?)
	 * - enable/disable SideBand usage
	 * - set 4x/2x/1x speed
	 * - enables AGP!
	 */
	softstate->am_enabled=1;

	softstate->dev_dma_attr=&dev_access;

	return 0;
}


/* handles AGPIOC_ALLOC
 *
 * Given a number of pages, allocate that number of pages.
 * copy out the beginning physical address, and a 'key' to
 * identify the memory for a bind call.
 *
 * Caller may use return value as return for ioctl(9e)
 */
int ioctl_i440bx_alloc(struct agpgart_softstate *softstate, void *arg,
		       int flags)
{
	agp_allocate alloc_info;
	uint32_t key;
	size_t length;
	struct keytable_ent *keyentry;

	if(ddi_copyin(arg, &alloc_info, sizeof(agp_allocate),flags) !=0){
		return (EFAULT);
	}
	length=AGP_PAGESTOBYTES(alloc_info.pg_count);

	key=agp_alloc_mem(softstate,length);
	if(key==-1){
		return (ENOMEM);
	}
	keyentry=agp_get_keyent(softstate,key);

	alloc_info.key=key;
	alloc_info.physical=keyentry->phys_addr;


	if(ddi_copyout(&alloc_info, arg,
		       sizeof(agp_allocate), flags) !=0){
		return (EFAULT);
	}
	
	return 0;
}
/* handles AGPIOC_DEALLOC
 *
 * This entry point is UNIQUE in that it is the only one to have a pure
 * int value as the arg. All the rest have (pointer to struct)
 *
 * Key referenced in arg must have already been unbound with AGPIOC_UNBIND
 *
 * Caller may use return value as return for ioctl(9e)
 * 0 is successfull dealloc
 */
int ioctl_i440bx_dealloc(struct agpgart_softstate *softstate, void *arg,
		       int flags)
{
	int key =(int)arg;
	int retval=0, delstat;
	agpgart_debugprint("ioctl_i440bx_dealloc: Key is %d\n",key);
	
	delstat=agp_dealloc_mem(softstate,key);

	if(delstat !=0){
		retval=EINVAL;
	}
	
	return retval;
}

/* handles AGPIOC_BIND:
 *    Given a keyID, and agp pagenumber, bind the memory tired
 *  to that key, to the agp memory starting at that agp pagenumber.
 *
 * returns 0 on okay, or appropriate errno otherwise.
 * Caller may use return value as return for ioctl(9e)
 *
 * NOTE: THis routine is 100% non-hardware specific.
 * When porting to other AGP hardware, you can probably
 * copy this routine verbatim.
 */
int ioctl_i440bx_bind(struct agpgart_softstate *softstate, void *arg,
		       int flags){
	int key, retval=0;
	agp_bind bindinfo;
	struct keytable_ent *keyent;
	uint32_t agp_address;

	if(ddi_copyin(arg, &bindinfo, sizeof(bindinfo),flags) !=0){
		return EFAULT;
	}
	key=bindinfo.key;
	agp_address=AGP_PAGESTOBYTES(bindinfo.pg_start);
	agpgart_debugprint("DEBUG: i440bx_bind called: key is %d, agp addr is 0x%x(page %d)\n",
			   key, agp_address, bindinfo.pg_start);

	keyent=agp_get_keyent(softstate,key);
	if(keyent==NULL){
		return EINVAL;
	}
	if(keyent->am_bound != 0){
		agpgart_debugprint("DEBUG: agp key %d already bound\n",key);
		agpgart_debugprint("   (am_bound=%d)\n",keyent->am_bound);
		return EINVAL;
	}

	retval=i440bx_add_to_gatt(softstate, keyent, agp_address);

	if(retval!=0){
		agpgart_debugprint("DEBUG: i440bx_add_to_gatt returned %d\n",retval);
	} else {
		keyent->agp_offset=agp_address;
		keyent->am_bound=1;
		agpgart_debugprint("DEBUG: ioctl_i440bx_bind: bound %d@0x%x\n",
			   key, agp_address);
	}
	
	return retval;
}


/* handles AGPIOC_UNBIND:
 *    Given a keyID, UN-bind the memory tied
 *  to that key from the GATT table.
 * Also unbinds the dma handle, but does NOT deallocate it, in case
 * user wants to rebind the memory later.
 * (Not sure why this should be neccessary, but linux supports it, so
 *   we kinda have to also)
 *
 * returns 0 on okay, or appropriate errno otherwise.
 * Caller may use return value as return for ioctl(9e)
 *
 * NOTE: THis routine is 100% non-hardware specific.
 * When porting to other AGP hardware, you can probably
 * copy this routine verbatim.
 */
int ioctl_i440bx_unbind(struct agpgart_softstate *softstate, void *arg,
		       int flags){
	int key, retval=0;
	agp_unbind unbindinfo;
	struct keytable_ent *keyent;

	if(ddi_copyin(arg, &unbindinfo, sizeof(unbindinfo),flags) !=0){
		return EFAULT;
	}
	key=unbindinfo.key;
	agpgart_debugprint("DEBUG: i440bx_unbind: key is %d\n", key);

	keyent=agp_get_keyent(softstate,key);
	if(keyent==NULL){
		return EINVAL;
	}
	if(keyent->am_bound == 0){
		agpgart_debugprint("DEBUG: i440bx_unbind: agp key %d not bound\n",key);
		return EINVAL;
	}
	retval=i440bx_remove_from_gatt(softstate,
			       keyent->agp_offset,
			       keyent->length);
	if(retval!=0){
		agpgart_debugprint("DEBUG: agp_remove_from_gatt returned %d\n",retval);
	} else {
		ddi_dma_unbind_handle(keyent->dma_handle);
		keyent->agp_offset=0;
		keyent->am_bound=0;
	}
			
	return retval;
}
