/*
 * agpgart.c: Main file for agpgart driver for Solaris
 * Written by Philip Brown : http://www.bolthole.com/solaris/
 *
 * This was grown out of my PCI "base", at
 * http://www.bolthole.com/solaris/drivers/
 *
 * This file contains the Solaris standard kernel driver entry points,
 * as well as hardware-neutral routines.
 * Each supported type of AGP hardware has its own specific
 * source code file module. eg: agp_i440bx.c
 *
 * See LICENSE file for copying/modification information
 *
 * %A%
 */

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/file.h>	/* for FKIOCTL */

/* THESE MUST BE THE LAST of the system includes */
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "agpgart.h"
#include "agpdrv.h"
#include "agp_i440bx.h"



/* This first set of defines and stuff is left over from my "base"
 * driver. It mostly probably wont get used much, but oh well.
 */

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)

/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"agpgart"	
#define DEVNODE_TYPE	DDI_NT_SERIAL_MB /* got a better type? */
 /* see sunddi.h for alternate types*/

/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	agpgart_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	agpgart_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
#ifdef OLD_MMAP
static	int	agpgart_mmap(dev_t dev, off_t offset, int prot);
#endif

static	int	agpgart_devmap(dev_t dev, devmap_cookie_t, offset_t, size_t,
		               size_t*, uint_t);


static	int	agpgart_open(dev_t *dev, int openflags, int otyp, cred_t *credp);
static	int	agpgart_print(dev_t, char*);
static	int	agpgart_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	agpgart_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	agpgart_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	agpgart_cb_ops = {
	agpgart_open,		/* open() */
	agpgart_close,		/* close() */
	nodev,			/* strategy()		*/
	agpgart_print,		/* print routine	*/
	nodev,			/* no dump routine	*/
	nodev,			/* read() */
	nodev,			/* write() */
	agpgart_ioctl,		/* ioctl */
	agpgart_devmap,		/*  devmap routine	*/
	nodev,			/* no longer use mmap routine */
	ddi_devmap_segmap,	/* system segmap routine */
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP,		/* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops agpgart_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	agpgart_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	nulldev,		/* devo_probe */
	agpgart_attach,		/* devo_attach */
	agpgart_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&agpgart_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"AGP driver v0.1",
	&agpgart_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 2 /* max of 1 hardware devices loaded */
			/* the naming scheme in /dev wont take more */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   agpgart_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * agpgart_glob_soft_handle;

/*************************************************************/
/* Utils here */

/* translate value read from pci reg CONF_APSIZE */
size_t agp_apsize_to_bytes(uint8_t apsize){
	switch(apsize &0x3f){
		case 0x3f: return 0x0400000;  /* 4 megs */
		case 0x3e: return 0x0800000;  /* 8 megs */
		case 0x3c: return 0x1000000;  /* 16 megs */
		case 0x38: return 0x2000000;  /* 32 megs */
		case 0x30: return 0x4000000;  /* 64 megs */
		case 0x20: return 0x8000000;  /* 128 megs */
		case 0x00:return 0x10000000;  /* 256 megs */
		default:
			cmn_err(CE_WARN,"unrecognized apsize value");
			return 0x04000000;
			
	}
}

/* translate value read from pci reg CONF_APSIZE */
/* See also apsize_to_bytes */
int agp_apsize_to_megs(uint8_t apsize){
	switch(apsize &0x3f){
		case 0x3f: return 0x04;  /* 4 megs */
		case 0x3e: return 0x08;  /* 8 megs */
		case 0x3c: return 0x10;  /* 16 megs */
		case 0x38: return 0x20;  /* 32 megs */
		case 0x30: return 0x40;  /* 64 megs */
		case 0x20: return 0x80;  /* 128 megs */
		case 0x00: return 256;
		default:
			cmn_err(CE_WARN,"unrecognized apsize value");
			return (0x04);
			
	}
}

/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
agpgart_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	va_start(args);
	fmt = va_arg(args,char*);

	if(debug_level==0){
		return;
	}

	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the agpgart_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See agpgart_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	agpgart_debugprint("agpgart: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&agpgart_global_mutex, "agpgart: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct agpgart_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&agpgart_glob_soft_handle,
				sizeof(struct agpgart_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&agpgart_global_mutex);
		ddi_soft_state_fini(&agpgart_glob_soft_handle);
		cmn_err(CE_NOTE, "agpgart: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	cmn_err(CE_NOTE, "agpgart: _init returning SUCCESS, supposedly\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	cmn_err(CE_NOTE, "agpgart: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	cmn_err(CE_NOTE, "agpgart: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&agpgart_glob_soft_handle);
		mutex_destroy(&agpgart_global_mutex);
	}
	return (ret);
}

/************************************************************/

/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS
 */
static int agpgart_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "agpgart-print(%d): %s",instance,string);
	return 0;
}

static int	agpgart_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	agpgart_debugprint("agpgart: getinfo called\n");
	return DDI_FAILURE;
}



/**********************************************************************/
/*       Key handling section starts here                             */
/*                                                                    */
/* Note: I deliberately do not have these routines automatically      */
/*  grab the instance_mutex, because that gets in the way of cleanup  */
/* routines holding the mutex across multiple routines.               */
/* Having to hold, release, hold the mutex in that kind of situation  */
/* is something I want to avoid.                                      */
/* 'Course, I suppose I could always have a SEPARATE mutex for keys...*/
/* but maybe theres something positive about having a corse mutex.    */
/**********************************************************************/


/* agp_add_keyent():
 *   called by agp_alloc_mem, and probably should not be called by
 *   anything else.
 * Adds an entry to our key table. Does NOT mess with anything else.
 * Assume this is a part of handling AGPIOC_ALLOC,
 * but actually only gets called directly by agp_alloc_mem()
 *
 * Caller needs to set values->dma_handle, acc_handle, length, real_length
 * MAKE SURE EVERYTHING ELSE IS SET TO ZERO, or you screw things up!!!
 *
 * returns key value 0 - AGP_MAXKEYS, or -1 on error.
 *    (error like no more keys free)
 * 
 * MUTEXES: none. Caller should choose their own mutex to use.
 *  Although you could just use instance_mutex.
 */
int agp_add_keyent(struct agpgart_softstate *softstate,
	       struct keytable_ent*values){
	struct keytable_ent *entry;
	int key;

	for(key=0;key<AGP_MAXKEYS; key++){
		entry=&softstate->keytable[key];
		if(entry->dma_handle==NULL){
			break;
		}
	}
	if(key>=AGP_MAXKEYS){
		mutex_exit(&softstate->instance_mutex);
		return -1;
	}

	agpgart_debugprint("agp_add_keyent: adding key %d\n",key);

	bcopy(values,entry,sizeof(struct keytable_ent));

	return key;
}

/* Used by AGPIOC_BIND, UNBIND, and DEALLOCATE related routines.
 * Given a key index (the user-visible id), find matching info struct.
 * If key not found, return NULL
 * Almost NO checking of validity of data. Just key range check right now.
 *
 * MUTEXES: none. Caller should choose their own mutex to use.
 *  Although you could just use instance_mutex.
 */
struct keytable_ent* agp_get_keyent(struct agpgart_softstate *softstate,
				    int key){
	if(key>=AGP_MAXKEYS){
		agpgart_debugprint("DEBUG: agp_get_keyent passed large key\n");
		return NULL;
	}
	if(key<0){
		agpgart_debugprint("DEBUG: agp_get_keyent passed bad key\n");
		return NULL;
	}
	return &softstate->keytable[key];
}

/* Look up the appropriate entry in our agp memory table, that
 * matches the given offset in the agp memory window
 * If AGPIOC_BIND has not been called for a range spanning the offset,
 * there will be nothing to return.
 *
 * returns NULL on out-of-range, or no entry.
 *
 * MUTEXES: none. But whatever you use for keyents, use for this
 *  routine also. Note: Calls agp_find_keyent;
 */
struct keytable_ent* 
agp_find_keyent(struct agpgart_softstate *softstate,
					  uint32_t req_offset)
{
	int keycount;
	struct keytable_ent *entry;

	agpgart_debugprint("find_keyent looking for offset 0x%x\n",req_offset);

	for(keycount=0;keycount<AGP_MAXKEYS;keycount++){
		entry=agp_get_keyent(softstate,keycount);
		if(entry->am_bound==0){
			agpgart_debugprint("Key %d not bound:skipping\n",
					   keycount);
			continue;
		}
		agpgart_debugprint("Key %d bound to offset 0x%x, len 0x%x\n",
				   keycount, entry->agp_offset,
				   entry->length);

		if(req_offset < entry->agp_offset)
			continue;
		if(req_offset >= (entry->agp_offset + entry->length))
			continue;

		
		agpgart_debugprint("Returning entry for Key %d\n",keycount);
		return entry;
	}

	return NULL;
}


#ifdef OLD_MMAP
/* Used by agpgart_mmap(). Nothing else should have to.
 * Which means that technically, this is an obsolete routine, and
 * could be removed, but I'm leaving it in here, "just in case"
 *
 * First look in keytable to find a keyed entry that
 * matches the agp window offset we have been given
 *
 * Then calculate the kernel-space address that we have bound to
 * that region.
 *
 * Returns the kernel-space address on success,
 *  otherwise, returns NULL
 *
 * MUTEXES: none. But whatever you use for keyents, use for this
 *  routine also. Note: Calls agp_find_keyent;
 */
static caddr_t agp_get_boundaddr(struct agpgart_softstate *softstate,
					  uint32_t req_offset)
{
	caddr_t returnaddr;
	struct keytable_ent *entry;

	/*
	agpgart_debugprint("agp_get_boundaddr: looking for %x\n",
			   req_offset);
	*/

	entry=agp_find_keyent(softstate,req_offset);
	if(entry==NULL) return NULL;
	
	/*kern_addr is where entry->agp_offset is mapped to */
	returnaddr=entry->kern_addr - entry->agp_offset + req_offset;

	return returnaddr;
}
#endif /* OLD_MMAP */


/*
 * This should only be called by agp_dealloc_mem()
 *  [and agp_del_allkeys() ]
 *
 * Removes an entry to our key table. Does NOT mess with anything else.
 * NULLs out dma_handle to mark entry as free.
 * Does NOT do any actual memory freeing, handle unbinding, etc.
 * That is done by agp_dealloc_mem()
 *
 * Note: "key" is simply the index into the keytable.
 * Return 0 on okay,
 * or some errno if you gave a bad key,
 *    OR if key is still bound to a GATT entry.
 * (You need to call the UNBIND routine for the key before calling this func.)
 *
 * MUTEXES:
 *    none. Up to caller what mutex to use. Make sure you are consistent
 *    about using the same mutex for any and all key accesses.
 *    Easy way out: Just use instance_mutex.
 */
static int agp_del_keyent(struct agpgart_softstate *softstate,int key){
	struct keytable_ent *entry;
	if(softstate==NULL){
		agpgart_debugprint("agp_del_keyent: no softstate???\n");
		return EFAULT;
	}
	agpgart_debugprint("agp_del_keyent: asked to delete key %d\n",key);

	entry=&softstate->keytable[key];
	if(entry->dma_handle==NULL){
		cmn_err(CE_WARN,"del_keyent: ERROR: bad key value");
		return EINVAL;
	}

	if(entry->am_bound){
		/* "bound" to GATT table. Not to be confused with
		 * "is dma handle bound"
		 */
		agpgart_debugprint("ERROR: del_keyent: key %d bound\n");
		return EINVAL;
	}
	bzero(entry,sizeof(struct keytable_ent));


	return 0;
}

/* walk through the key table, and count the number of pages
 * accounted for in the keytable.
 * This is a pure "allocated" count. It does not care about
 * bound vs unbound.
 * MUTEXES: none
 *  If called from agpgart.c, caller should hold instance_mutex.
 *  If from other hardware-specific file, mutex is already
 *   held for you, so dont worry about it.
 */
int agp_count_keypages(struct agpgart_softstate *softstate){
	int key;
	int pagecount=0;

	for(key=0;key<AGP_MAXKEYS; key++){
		int klength=softstate->keytable[key].length;
		if(klength!=0){
			pagecount+=AGP_BYTESTOPAGES(klength);
		}
	}

	return pagecount;
}

/* Walk the keytable one by one, and delete all the keys.
 * Uses del_keyent()
 * MUTEXES: See comments for del_keyent()
 */
void agp_del_allkeys(struct agpgart_softstate *softstate){
	int key;
	agpgart_debugprint("agp_del_allkeys called\n");
	for(key=0;key<AGP_MAXKEYS;key++){
		if(softstate->keytable[key].dma_handle !=NULL){
			agp_del_keyent(softstate,key);
		}
	}
	agpgart_debugprint("agp_del_allkeys done\n");
}


/**********************************************************************/
/*                    END of key handling section                     */
/**********************************************************************/


/* Shared routine that implements the heart of AGPIOC_ALLOC
 * First allocates memory suitable for using with the AGP hardware.
 * Then inserts alloc info into keytable, via agp_add_keyent()
 * and  returns the key value on success.
 * Otherwise, returns -1 on fail.
 *
 * NOTE: caller gets the physical address of mem by looking it up via the key.
 *       caller can then return the physaddr info that the AGPIOC_ALLOC
 *       ioctl likes to have.
 *
 * MUTEXES:
 *   Use the same mutex you use to protect any key access.
 *   You can wimp out and just use instance_mutex if you like.
 */
int agp_alloc_mem(struct agpgart_softstate *softstate, size_t length){
	struct keytable_ent keyentry;
	caddr_t newmem;
	int bind_val, key_val;

	/* Clean out our temp struct on stack first. In userland,
	 * stack is zeroed, but in kernel, it apparently is NOT!
	 */
	bzero(&keyentry,sizeof(struct keytable_ent));

	keyentry.length=length;

	agpgart_debugprint("agp_alloc_mem called with length %d\n",length);

	if(ddi_dma_alloc_handle(softstate->dip,
				&dev_dma_attr,
				DDI_DMA_SLEEP, NULL,
				&keyentry.dma_handle) != DDI_SUCCESS)
	{
		return -1;
	}

	newmem=ddi_umem_alloc(length, DDI_UMEM_SLEEP,&keyentry.umem_cookie);
	if(newmem==NULL){
		ddi_dma_free_handle(&keyentry.dma_handle);
		return -1;
	}
	keyentry.kern_addr=newmem;

/******** Maybe this is redundant, if I'm also using ddi_dma_addr_bind_handle */
#ifdef UMEM_LOCK
	if(ddi_umem_lock(newmem, length,
			 DDI_UMEMLOCK_READ|DDI_UMEMLOCK_WRITE,
			 &keyentry.umem_cookie) != 0){
		agpgart_debugprint("Could not lock umem\n");
		ddi_dma_free_handle(&keyentry.dma_handle);
		ddi_umem_free(keyentry.umem_cookie);
	}
#endif /* UMEM_LOCK */

	/* Would LIKE to do this in the _bind routines.
	 * But this is the only clean way to get the starting
	 * physical address.
	 * Even though I dont support returning it yet, I am
	 * keeping this structure for when and if I DO decide to
	 * support it.
	 */
	bind_val=ddi_dma_addr_bind_handle(keyentry.dma_handle,
				    NULL,
				    newmem, length,
				    DDI_DMA_RDWR,
				    DDI_DMA_SLEEP,NULL,
				    &keyentry.dma_cookie,
				    &keyentry.num_dma_cookies);

	if(bind_val != DDI_DMA_MAPPED){
		agpgart_debugprint("bind failed: bindval=%d, cookiecount=%d\n",
				   bind_val, keyentry.num_dma_cookies);
		agpgart_debugprint(" kernel addr of newmem = %x\n",
				   newmem);

		ddi_dma_unbind_handle(keyentry.dma_handle);
		ddi_dma_free_handle(&keyentry.dma_handle);

#ifdef UMEM_LOCK
		ddi_umem_unlock(keyentry.umem_cookie);
#endif
		ddi_umem_free(keyentry.umem_cookie);
		return -1;
	}

#ifdef RETURN_PHYSADDR
	/* we do NOT set this normally. only for special hardwarespecific
	 * alloc "types".
	 * Which would require that the entire alloced chunk be
	 * contiguous, also. Forget that.
	 */ 
	keyentry.phys_addr=keyentry.dma_cookie.dmac_address;
#endif

	key_val=agp_add_keyent(softstate,&keyentry);
	if(key_val==-1){
		ddi_dma_unbind_handle(keyentry.dma_handle);
		ddi_dma_free_handle(&keyentry.dma_handle);
#ifdef UMEM_LOCK
		ddi_umem_unlock(keyentry.umem_cookie);
#endif
		ddi_umem_free(keyentry.umem_cookie);
	}

	agpgart_debugprint("alloc_mem: successfully allocated %d bytes for key %d, umemcookie %p\n",
			   length, key_val, &keyentry.dma_cookie);
	return(key_val);
				    
				    
}

/*
 * Hardware-neutral side of AGPIOC_DEALLOC
 * Unbinds handles and frees memory assigned by AGPIOC_ALLOC.
 * Clears "key" entry.
 * return 0 on okay, or EINVAL or something if invalid key.
 */
int agp_dealloc_mem(struct agpgart_softstate *softstate, int key){
	struct keytable_ent *entry;
	entry=agp_get_keyent(softstate,key);
	if(entry==NULL){
		agpgart_debugprint("agp_dealloc_mem: bad key %d\n",key);
		return EINVAL;
	}
	if(entry->dma_handle==NULL){
		agpgart_debugprint("agp_dealloc_mem: key %d orphaned\n",key);
		return EINVAL;
	}

	/* I really do NOT want to auto-unbind.
	 * But it seems that Jeff Hartmann wrote code taht
	 * assumed this behaviour, so I guess I have to support it.
	 */
	if(entry->am_bound){
		agp_unbind unbindinfo;
		int unbindval;
		unbindinfo.key=key;
		/* FKIOCTL: tell routine struct is in kernel space, NOT user */
		unbindval=ioctl_i440bx_unbind(softstate,&unbindinfo,FKIOCTL);
		if(unbindval!=0){
			agpgart_debugprint("agp_dealloc_mem: unbind of %d had ERR\n",
				   key);
		}
		return unbindval;
	}

	ddi_dma_free_handle(&entry->dma_handle);
#ifdef UMEM_LOCK
	ddi_umem_unlock(entry->umem_cookie);
#endif
	ddi_umem_free(entry->umem_cookie);

	return (agp_del_keyent(softstate,key));
}

/* This routine can be called by agpgart_attach (if errors occur)
 * and agpgart_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void agpgart_cleanup(struct agpgart_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: agpgart_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	if(softstate->intr_cookie != NULL){
		ddi_remove_intr(softstate->dip,0,softstate->intr_cookie);
	}
	if(softstate->pcihandle !=NULL){
		pci_config_teardown(&softstate->pcihandle);
	}

	if(softstate->keytable!=NULL){
		/* uses instance_mutex */
		kmem_free(softstate->keytable,
			  AGP_MAXKEYS*(sizeof (struct keytable_ent)) );
	}
	ddi_soft_state_free(agpgart_glob_soft_handle,instance);
	
}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in agpgart_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  agpgart_detach()
 */
static int agpgart_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance, status;
	struct agpgart_softstate *softstate;
	int tmpint;

	/* default debug level to **1**
	 * You can override in the .conf file with "debug-level=0".
	 * When you have a production-ready driver, change
	 * default to 0
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 1);

	agpgart_debugprint("agpgart: agpgart_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "agpgart: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "agpgart: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "agpgart: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	agpgart_debugprint("agpgart: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(agpgart_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "agpgart: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(agpgart_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "agpgart: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(agpgart_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;

	/* you will probably want this at some point or another */
	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);



	status = pci_config_setup(softstate->dip,&softstate->pcihandle);

	if (status != DDI_SUCCESS) {
		cmn_err(CE_WARN,
			 "could not map PCI regs");
			agpgart_cleanup(softstate);
			return DDI_FAILURE;
	}



	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */
	 
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* We want to do interrupt registering late in the game.
	 * Make sure everything is set up fully,
	 * just in case a random interrupt triggers the handler before
	 * we've even fully finished attach() !
	 */
	if(ddi_dev_nintrs(dip, &tmpint) == DDI_FAILURE){
		tmpint=0;
	}
	if(tmpint>0){
		cmn_err(CE_WARN,
			"WARNING: interrupt available.We do not use it");
	}

	softstate->keytable=kmem_zalloc(
				AGP_MAXKEYS*(sizeof (struct keytable_ent)),
				KM_SLEEP);
					
	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}

static int agpgart_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	struct agpgart_softstate *softstate;

   	agpgart_debugprint("agpgart: open called\n");

	/* even if you dont USE this, it is a check to see if we need
	 * to return ENXIO
	 */
	softstate=ddi_get_soft_state(agpgart_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is technically NOT an error; it is expected.
		 * We return ENXIO to support "deferred attach"
		 */
		return (ENXIO);
	}
	mutex_enter(&softstate->instance_mutex);
	if(softstate->am_open !=0){
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
		
	}
	softstate->am_open=1;
	mutex_exit(&softstate->instance_mutex);

	/* return 0 on okay, or an errno otherwise */
   	return 0;
}



static int agpgart_close(dev_t dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(dev);
	struct agpgart_softstate *softstate;

   	agpgart_debugprint("agpgart: close called\n");

	/* even if you dont USE this, it is a check to see if we need
	 * to return ENXIO
	 */
	softstate=ddi_get_soft_state(agpgart_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"agpgart_close: could not get softstate");
		return (ENXIO);
	}
	mutex_enter(&softstate->instance_mutex);

	if(softstate->am_open==0){
		mutex_exit(&softstate->instance_mutex);
		return(EINVAL);
	}
	softstate->am_open=0;

	mutex_exit(&softstate->instance_mutex);

	if(softstate->am_enabled==1){
		ioctl_i440bx_release(softstate,NULL,0);
	}
   	return (0);
}

static int agpgart_ioctl(dev_t dev, int cmd, intptr_t intarg, int flags,
		    cred_t *credp, int *rvalp){
   	int instance;
	int retval=ENXIO;
	void *arg=(void*)intarg;

	struct agpgart_softstate *softstate;
   	instance=DEVT_TO_INST(dev);
	softstate=ddi_get_soft_state(agpgart_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"agpgart_ioctl: could not get softstate");
		return -1;
	}

	/*agpgart_debugprint("agpgart: ioctl 0x%x called\n",(unsigned int)cmd);*/

	mutex_enter(&softstate->instance_mutex);

	switch(cmd){
	    case AGPIOC_INFO:
		retval=ioctl_i440bx_info(softstate, arg, flags);
		break;
	    case AGPIOC_ACQUIRE:
		retval=ioctl_i440bx_acquire(softstate, arg, flags);
		break;
	    case AGPIOC_RELEASE:
		retval=ioctl_i440bx_release(softstate, arg, flags);
		break;
	    case AGPIOC_SETUP:
		retval=ioctl_i440bx_setup(softstate, arg, flags);
		break;
	    case AGPIOC_ALLOCATE:
		retval=ioctl_i440bx_alloc(softstate, arg, flags);
		break;
	    case AGPIOC_DEALLOCATE:
		retval=ioctl_i440bx_dealloc(softstate, arg, flags);
		break;
	    case AGPIOC_BIND:
		retval=ioctl_i440bx_bind(softstate, arg, flags);
		break;
	    case AGPIOC_UNBIND:
		retval=ioctl_i440bx_unbind(softstate, arg, flags);
		break;
	    default:
		agpgart_debugprint("Error: unsupported ioctl %d\n",
				   cmd);
		agpgart_debugprint("   (AGPIOC_BASE=%d\n",AGPIOC_BASE);
		break;

	}
	mutex_exit(&softstate->instance_mutex);
	/*agpgart_debugprint("ioctl 0x%x complete\n",(unsigned int)cmd);*/
   	return (retval);
}

#ifdef OLD_MMAP
/*
 * This is no longer used. But I keep it in here as a teaching
 * excercise. See agpgart_devmap()  instead.
 *
 *
 * A few things to note:
 *   - The AGP user program has to have called AGPIOC_ALLOCATE **and* BIND
 *      memory, or else there is nothing to mmap
 *   - This is a legacy solaris entrypoint
 *   - This maps a SINGLE PAGE at a time.
 *      So be careful putting debug stuff in here!
 */
static int agpgart_mmap(dev_t dev, off_t offset, int prot){
   	int instance;
	struct agpgart_softstate *softstate;
	caddr_t kern_addr;
	ppid_t returnpage;

   	instance=DEVT_TO_INST(dev);
	softstate=ddi_get_soft_state(agpgart_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"agpgart_mmap: could not get softstate");
		return (NOPAGE);
	}

	mutex_enter(&softstate->instance_mutex);

	if(offset==0){
		agpgart_debugprint("minidebug: agpgart_mmap called for page0\n");
	}

	if(softstate->GATT==NULL){
		agpgart_debugprint("agpgart_mmap: no GATT allocated\n");
		mutex_exit(&softstate->instance_mutex);
		return (NOPAGE);
	}
	if(offset > softstate->aperture_size){
		mutex_exit(&softstate->instance_mutex);
		agpgart_debugprint("agpgart_mmap: offset %d too large\n",
				   offset);
		return (NOPAGE);
	}
	kern_addr=agp_get_boundaddr(softstate,offset);
	if(kern_addr==NULL){
		mutex_exit(&softstate->instance_mutex);
		agpgart_debugprint(
			"agpgart_mmap: agp_get_boundaddr failed for offset %x\n",
			offset);

		return (NOPAGE);
	}

	/* This is wierd, non-DDK, and is *EXACTLY* what is
	 * specified in mmap(9e)
	 * except hat_getfpfnum is now obsolete, and ddi.h
	 * mentions that kvtoppid is the replacement for it
	 */
	returnpage=kvtoppid(kern_addr);

	/*agpgart_debugprint("MMAP: returning page id %x for offset %d!!\n",
		       returnpage,offset);
	*/
	mutex_exit(&softstate->instance_mutex);
	if(offset==0){
		agpgart_debugprint("agpgart_mmap: returning pageid %x\n",
				   returnpage);
	}

	return(returnpage);
}

#endif /* OLD_MMAP */



/*
 *
 * This actually handles all mmap() calls.
 *
 * A few things to note:
 *   - The AGP user program has to have called AGPIOC_ALLOCATE **and* BIND
 *      memory, or else there is nothing to mmap
 *     (Plus AGPIOC_SETUP must have been called once)
 */
static int agpgart_devmap(dev_t dev, devmap_cookie_t cookie, offset_t offset,
                      size_t len, size_t *mappedlen, uint_t model)
{
	struct agpgart_softstate *softstate;
   	int instance, status;
	struct keytable_ent * mementry;
	offset_t local_offset;
	
   	instance=DEVT_TO_INST(dev);
	softstate=ddi_get_soft_state(agpgart_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"agpgart_devmap: could not get softstate");
		return (NOPAGE);
	}

	len=ptob(btopr(len)); /* make len an even pagesize */

	mutex_enter(&softstate->instance_mutex);

	if(offset==0){
		agpgart_debugprint("minidebug: agpgart_devmap called for page0\n");
	}

	if(softstate->GATT==NULL){
		agpgart_debugprint("agpgart_devmap: no GATT allocated\n");
		mutex_exit(&softstate->instance_mutex);
		return (NOPAGE);
	}
	if(offset > softstate->aperture_size){
		agpgart_debugprint("agpgart_devmap: offset %d too large\n",
				   offset);
		mutex_exit(&softstate->instance_mutex);
		return (NOPAGE);
	}

	mementry=agp_find_keyent(softstate, (offset & 0xffffffff));
	if(mementry==NULL){
		agpgart_debugprint("agpgart_devmap: no entry for offset %d\n",
				   offset);
		mutex_exit(&softstate->instance_mutex);
		return (NOPAGE);
	}
	if(softstate->dev_dma_attr==NULL){
		agpgart_debugprint("agpgart_devmap: dev_dma_attr is NULL\n");
		mutex_exit(&softstate->instance_mutex);
		return (NOPAGE);
	}

	local_offset=offset - mementry->agp_offset;
	agpgart_debugprint("devmap using local_offset=%lld, len=%ld\n",
	                   local_offset, len);

	if(len > mementry->length){
		len=mementry->length;
		/* It's okay to just truncate the length, as
		 * long as we tell the system we are truncating.
		 * We will then get called again with the remainder
		 */
	}

	agpgart_debugprint("using umem cookie %p\n",mementry->umem_cookie);

	status=devmap_umem_setup(cookie, softstate->dip, NULL,
	                           mementry->umem_cookie, local_offset, len,
	                           PROT_ALL, DEVMAP_DEFAULTS, softstate->dev_dma_attr);

	if(status==0){
		*mappedlen=len;
	} else {
		*mappedlen=0;
	}

	mutex_exit(&softstate->instance_mutex);

	agpgart_debugprint("agpgart_devmap: return status=%d\n",status);

	return (status);
}

static int agpgart_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct agpgart_softstate *softstate;
   	instance=ddi_get_instance(dip);

	agpgart_debugprint("agpgart: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(agpgart_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_NOTE,"agpgart_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	agpgart_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}

