#ifndef _AGPGART_H_
#define _AGPGART_H_

/***********************************************************************
 * This file documents what userland needs to access and use
 * /dev/agpgart
 *
 * This mostly mirrors linux naming, since we're trying to support
 * the originallly Linux API.
 * However, some types may been changed, sometimes out of neccesity,
 * and sometimes, because the linux ones just dont make sense.
 *
 * AGP is mostly used as a way for the graphics chips to quickly
 * access main RAM.
 * After you have gone through the ACQUIRE and SETUP ioctls,
 * you allocate some RAM through the AGPIOC_ALLOCATE ioctl.
 * If it succeeds, then AGPIOC_BIND it.
 * Once you have bound the ram, the card can access it.
 * For YOU to access it, you need to call mmap()
 * If the available aperture is 64 megs, and you want to map the entire
 * allocated RAM to you, you would call
 * char *agpmem=
 *   mmap((caddr_t)0, 64*1024*1024, PROD_READ|PROT_WRITE,MAP_SHARED,0);
 *
 *  and then you can access agpmem directly.
 *
 * Solaris version of agpgart by:
 *   Philip Brown, phil@bolthole.com
 * @(#) agpgart.h 1.5@(#)
 *
 ***********************************************************************
 */

/* You must have included <sys/types.h> before this file. */

/* These conversion routines are NOT in the linux agpgart.h currently.
 * -- Phil
 */
#define AGP_PAGE_SIZE	4096	/* 4k page size for AGP */
#define AGP_MEGSTOPAGES(x)	((x) << 8) /* number of pages in a megabyte */
#define AGP_PAGESTOBYTES(x)	((x) << 12)
#define AGP_BYTESTOPAGES(x)	((x) >> 12) /* page num for byte offset.
					  * assumes offset is on page boundry.
					  * (TRUNCATES!)
					  */
#define AGP_BYTESTOMEGS(x)	((x) >> 20)

/*
 * example: 
 * If(AGP_NONALIGNED(x)){ puts("x is not page aligned"); }
 */
#define AGP_NONALIGNED(x)	((x) & 0x0fff)


#define AGPIOC_BASE       'A'
#define AGPIOC_INFO       _IOR (AGPIOC_BASE, 0, agp_info*)
#define AGPIOC_ACQUIRE    _IO  (AGPIOC_BASE, 1)
#define AGPIOC_RELEASE    _IO  (AGPIOC_BASE, 2)
	/* ACQUIRE and RELEASE are NOPs.
	 * The driver does the implied operation on open() and close()
	 */
#define AGPIOC_SETUP      _IOW (AGPIOC_BASE, 3, agp_setup)
#define AGPIOC_RESERVE    _IOW (AGPIOC_BASE, 4, agp_region)
#define AGPIOC_PROTECT    _IOW (AGPIOC_BASE, 5, agp_region)
#define AGPIOC_ALLOCATE   _IOWR(AGPIOC_BASE, 6, agp_allocate)
#define AGPIOC_DEALLOCATE _IOW (AGPIOC_BASE, 7, int)
#define AGPIOC_BIND       _IOW (AGPIOC_BASE, 8, agp_bind)
#define AGPIOC_UNBIND     _IOW (AGPIOC_BASE, 9, agp_unbind)

#define AGP_DEVICE      "/dev/agpgart"


typedef struct _agp_version {
	uint16_t major;
	uint16_t minor;
} agp_version;

typedef struct _agp_info {
	agp_version version;
	uint32_t bridge_id;	/* bridge vendor+device*/
	uint32_t agp_mode;	/* mode of bridge */
	paddr_t aper_base;	/* base of aperture*/
	size_t aper_size;
	int pg_total;	/* max number of pages in aperture */
	int pg_system;	/* same as pg_total, it seems */
	int pg_used;		/* NUMBER of currently used pages */
} agp_info;

typedef struct _agp_setup {
	uint32_t agp_mode; /* this generally is just the same value
			    * that the INFO call says is the agp_mode
			    */
} agp_setup;

/* Used as part of agp_region only */
typedef struct _agp_segment {
	off_t pg_start;	/*offset into memory relative to agp window ?*/
	size_t pg_count;
	int prot;	/* ignored */
} agp_segment;

/* Used in conjunction with AGPIOC_RESERVE and AGPIOC_PROTECT */
typedef struct _agp_region {
	pid_t pid;		/* ignored */
	size_t seg_count;	/* number of segments in seg_list */
	agp_segment *seg_list;
} agp_region;


/* I dont want to waste memory remembering more than 64 keys.
 * You shouldnt need more than 64 keys anyway: Most programs
 * only need **one**.
 * If you honestly have a need for more than 64, email me.
 */

#define AGP_MAXKEYS 64

/* for AGPIOC_ALLOCATE */
typedef struct _agp_allocate {
	uint32_t key;		/* This is returned by driver.
				 * Use it with AGPIOC_BIND.
				 * There IS an internal limit as to how
				 * many you can allocate.
				 * I am adding AGP_MAXKEYS even though LINUX
				 * doesnt have it public
				 */
	int pg_count;		/* how many AGP_PAGE_SIZE pages you want?*/
	uint32_t type;		/* leave as 0. ignored */
	paddr_t physical;	/* Return to caller what physical (I/O) memory
				 * the memory has been allocated at.
				 * THIS DOES NOT MEAN THE CALLER CAN
				 * WRITE DIRECTLY TO IT!!
				 * The device would have to, somehow.
				 */
} agp_allocate;


/* for AGPIOC_BIND:
 * Once you have ALLOCATEd memory, this will then bind it into the
 * "GATT" table, so the card can access it.
 */
typedef struct _agp_bind {
	int key;		/* value from AGPIOC_ALLOCATE call */
	int pg_start;		/* say AGP aperture page number you
				 * want it bound to
				 */
} agp_bind;

typedef struct _agp_unbind {
	int key;		/* value from AGPIOC_ALLOCATE call */
	uint32_t priority;	/* ignored
				 */
} agp_unbind;


#endif /*_AGPGART_H_*/
